#define _GNU_SOURCE

#include <error.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <gelf.h>
#include <sys/procfs.h>
#include <link.h>
#include <popt.h>
#include <inttypes.h>
#include <sys/mman.h>

#include "core_map.h"

int print_debug = 0;
static int print_kernel_buildid = 0;
static int print_vdso_buildid;

static const char *core_fname;
static const char *core_psargs;

static GElf_Addr core_l_addr;
static GElf_Addr core_vdso_l_addr;

/* This is the linkers shard object map */
static struct link_map *core_link_map = NULL;

static Elf *map_search_ehdr(struct core_map_info *map);

static void DEBUG(const char *format, ...)
{
	if (print_debug) {
		va_list alist;

		va_start(alist, format);
		fputs("DEBUG: ", stderr);
		vfprintf(stderr, format, alist);
		va_end(alist);
	}
}

static struct poptOption optionsTable[] = {
	{ "debug",  '\0', POPT_ARG_NONE, &print_debug, 0,
	  "print debug information", NULL },
	{ "kernel", '\0', POPT_ARG_NONE, &print_kernel_buildid, 0,
	  "print the build id of the current running kernel", NULL },
	{ "vdso", '\0', POPT_ARG_NONE, &print_vdso_buildid, 0,
	  "print the build id of the mapped vDSO", NULL },
	POPT_AUTOHELP
	{ NULL, 0, 0, NULL, 0, NULL, NULL }
};

static const char *elf_typenames[] = {
	[PT_NULL] = "NULL",
	[PT_LOAD] = "LOAD",
	[PT_DYNAMIC] = "DYNAMIC",
	[PT_INTERP] = "INTERP",
	[PT_NOTE] = "NOTE",
	[PT_SHLIB] = "SHLIB",
	[PT_PHDR] = "PHDR",
	[PT_TLS] = "TLS",
};
static const char *elf_classnames[] = {
	[ELFCLASS32] = "ELFCLASS32",
	[ELFCLASS64] = "ELFCLASS64",
};
static const char *elf_datanames[] = {
	[ELFDATA2LSB] = "ELFDATA2LSB",
	[ELFDATA2MSB] = "ELFDATA2MSB",
};

static void print_phdr(Elf *elf, GElf_Phdr *phdr, int cnt, int indent)
{
	char buf[19];
	const char *p_type;
	char space[20];

	if (indent > 0) {
		memset(space, ' ', indent);
		space[indent] = '\0';
	} else
		space[0] = '\0';

	if (phdr->p_type >= PT_NUM) {
		snprintf (buf, sizeof (buf), "%x", phdr->p_type);
		p_type = buf;
	} else
		p_type = elf_typenames[phdr->p_type];

	fprintf(stderr,
		"%s%3d %-7s %#0*llx %#0*llx %#0*llx %#0*llx %#0*llx %#6llx ",
		space,
		cnt, p_type,
		gelf_getclass (elf) == ELFCLASS32 ? 9 : 17,
		(unsigned long long int) phdr->p_offset,
		gelf_getclass (elf) == ELFCLASS32 ? 10 : 18,
		(unsigned long long int) phdr->p_vaddr,
		gelf_getclass (elf) == ELFCLASS32 ? 10 : 18,
		(unsigned long long int) phdr->p_paddr,
		gelf_getclass (elf) == ELFCLASS32 ? 9 : 12,
		(unsigned long long int) phdr->p_filesz,
		gelf_getclass (elf) == ELFCLASS32 ? 9 : 12,
		(unsigned long long int) phdr->p_memsz,
		(unsigned long long int) phdr->p_align);
	putc_unlocked ((phdr->p_flags & PF_X) ? 'X' : ' ', stderr);
	putc_unlocked ((phdr->p_flags & PF_W) ? 'W' : ' ', stderr);
	putc_unlocked ((phdr->p_flags & PF_R) ? 'R' : ' ', stderr);
	putc_unlocked ('\n', stderr);
}

static int nhdr_getbuildid(Elf_Data *data, unsigned char *dest,
			   size_t dest_len)
{
	GElf_Nhdr nhdr;
	size_t pos = 0, next, name_pos, desc_pos;
	int found = 0;

	if (data->d_type != ELF_T_NHDR)
		return -1;

	do {
#define NOTE_ALIGN(n) (((n) + 3) & -4U)
		GElf_Nhdr *n = data->d_buf + pos;
		name_pos = pos + sizeof(*n);
		desc_pos = name_pos + NOTE_ALIGN(n->n_namesz);
		next = desc_pos + NOTE_ALIGN(n->n_descsz);
		nhdr = *n;

		DEBUG("next=%#x type=%#x, namepos=%#x(%u), descpos=%#x(%u)\n",
		      next, nhdr.n_type, name_pos, nhdr.n_namesz, desc_pos,
		      nhdr.n_descsz);

		if (nhdr.n_type == NT_GNU_BUILD_ID
		    && nhdr.n_namesz == sizeof "GNU"
		    && !memcmp (data->d_buf + name_pos, "GNU", sizeof "GNU")) {
			const unsigned char *id = data->d_buf + desc_pos;

			DEBUG("Found build-id at %#x (%#x, %#x):",
			      pos, name_pos, desc_pos);
			if (print_debug) {
				unsigned int i;
				fprintf(stderr, ".note.gnu.build-id: ");
				for (i = 0; i < nhdr.n_descsz; i++)
					fprintf(stderr, "%02" PRIx8, id[i]);
				fprintf(stderr, "\n");
			}

			if (dest_len < nhdr.n_descsz)
				fprintf(stderr,
					"WARNING: build-id is truncated\n");
			if (found)
				fprintf(stderr,
					"ERROR: multiple build-id detected\n");
			memcpy(dest, id, dest_len < nhdr.n_descsz ?
			       dest_len : nhdr.n_descsz);
			found = dest_len < nhdr.n_descsz ? 0 : nhdr.n_descsz;
		}

		pos = next;
	} while (pos > 0 && pos < data->d_size);

	return found;
}

static int get_buildid(Elf *elf, GElf_Phdr *phdr, unsigned char *dest,
		       size_t dest_len)
{
	size_t pos = 0, next, name_pos, desc_pos;
	GElf_Nhdr nhdr;
	Elf_Data *data;
	int found = 0;

	if (phdr->p_type != PT_NOTE)
		return 0;

	data = elf_getdata_rawchunk (elf, phdr->p_offset, phdr->p_filesz,
				     ELF_T_NHDR);
	if (!data) {
		fprintf(stderr, "Couldn't read NOTE segment: %s\n",
			elf_errmsg(-1));
		return 0;
	}

	return nhdr_getbuildid(data, dest, dest_len);
}

static void so_print_buildid(struct core_map_info *core_map, const char *name,
			     GElf_Addr ehdr_addr)
{
	struct core_map_info *so_map_info;
	Elf *so_elf;
	GElf_Ehdr so_ehdr;

	so_map_info = find_map(core_map, ehdr_addr, ELF_T_EHDR);
	if (!so_map_info) {
		DEBUG("No map for address %#" PRIx64 "\n", ehdr_addr);
		return;
	}
	so_elf = map_search_ehdr(so_map_info);
	if (!so_elf) {
		DEBUG("No ELF magic found at address %#" PRIx64 "\n",
		      ehdr_addr);
		return;
	}

	if (gelf_getehdr (so_elf, &so_ehdr) == NULL) {
		fprintf(stderr, "Can't get the ELF header: %s\n",
			elf_errmsg (-1));
		return;
	}

	/* Look for NOTE segments that contain a build-id */
	int i;
	for (i = 0; i < so_ehdr.e_phnum; i++) {
		GElf_Phdr note_mem;
		GElf_Phdr *note_phdr = gelf_getphdr (so_elf, i, &note_mem);

		if (print_debug)
			print_phdr(so_elf, note_phdr, i, 2);

		if (note_phdr->p_type != PT_NOTE)
			continue;

		/* Read the build-id */
		unsigned char build_id[note_phdr->p_filesz];
		size_t len = get_buildid(so_elf, note_phdr, build_id,
					 note_phdr->p_filesz);
		if (len) {
			unsigned int i;
			printf("%s %#" PRIx64 " ", name, ehdr_addr);
			for (i = 0; i < len; i++)
				printf("%02" PRIx8, build_id[i]);
			printf("\n");
		}
	}
}

static int kernel_print_buildid(void)
{
	struct stat stat_buf;
	int fd, rc;
	void *note_mem = NULL;
	void *note_mmap;

	fd = open("/sys/kernel/notes", O_RDONLY);
	if (fd == -1) {
		error (0, errno, "cannot open notes file");
		return -1;
	}

	rc = fstat(fd, &stat_buf);
	if (rc) {
		error (0, errno, "cannot stat notes file");
		close(fd);
		return -1;
	}

	note_mmap = mmap(NULL, stat_buf.st_size, PROT_READ,
			 MAP_PRIVATE, fd, 0);
	if (note_mmap == MAP_FAILED) {
		note_mmap = note_mem = malloc(stat_buf.st_size);
		if (!note_mmap) {
			error (0, errno, "cannot allocate memory");
			close(fd);
			return -1;
		}

		rc = read(fd, note_mmap, stat_buf.st_size);
		if (rc != stat_buf.st_size) {
			error (0, errno, "cannot read notes file");
			close(fd);
			return -1;
		}
	}

	Elf_Data data_mem;
	Elf_Data *data = &data_mem;
	data_mem.d_buf = note_mmap;
	data_mem.d_size = stat_buf.st_size;
	data_mem.d_type = ELF_T_NHDR;
	data_mem.d_align = 1;
	data_mem.d_version = 1;

	unsigned char dest[stat_buf.st_size];
	size_t dest_len = stat_buf.st_size;

	rc = nhdr_getbuildid(data, dest, dest_len);
	if (rc > 1) {
		unsigned int i;
		printf("vmlinux ");
		for (i = 0; i < rc; i++)
			printf("%02" PRIx8, dest[i]);
		printf("\n");
	}

	if (note_mmap != note_mem)
		munmap(note_mmap, stat_buf.st_size);
	else
		free(note_mem);
	close(fd);
	return 0;
}

static Elf_Data *auxv = NULL;
static unsigned long auxv_size = 0;

static int core_handle_note(Elf *elf, GElf_Phdr *phdr)
{
	size_t pos = 0;
	GElf_Nhdr nhdr;
	size_t name_pos;
	size_t desc_pos;
	Elf_Data *notes;

	if (phdr->p_type != PT_NOTE)
		return 1;

	notes = elf_getdata_rawchunk(elf, phdr->p_offset, phdr->p_filesz,
				     ELF_T_NHDR);
	if (!notes)
		return 1;

	while ((pos = gelf_getnote (notes, pos, &nhdr, &name_pos, &desc_pos)) > 0) {
		DEBUG("Note header with n_type = %d and n_descsz = %u\n",
		      nhdr.n_type, nhdr.n_descsz);

		if (nhdr.n_namesz != sizeof "CORE" ||
		    memcmp (notes->d_buf + name_pos, "CORE", sizeof "CORE"))
			continue;

		switch (nhdr.n_type) {
		case NT_AUXV:
			auxv_size = nhdr.n_descsz;
			auxv = elf_getdata_rawchunk(elf,
						    phdr->p_offset + desc_pos,
						    nhdr.n_descsz,
						    ELF_T_AUXV);

			if (!auxv) {
				fprintf(stderr, "Can't get AUXV: %s\n",
					elf_errmsg(-1));
				return 1;
			}

			DEBUG("Found AUXV at %p with length 0x%x\n",
			      auxv, auxv_size);
			break;

		case NT_PRPSINFO:
			switch(nhdr.n_descsz) {
			case 124:
				core_fname = strndup(notes->d_buf + desc_pos +
						     28, 16);
				core_psargs = strndup(notes->d_buf + desc_pos +
						      44, 80);
				break;
			case 136:
				core_fname = strndup(notes->d_buf + desc_pos +
						     40, 16);
				core_psargs = strndup(notes->d_buf + desc_pos +
						      56, 80);
				break;
			}

#if 0
			/* Need to copy 32bit structs here */
			if (nhdr.n_descsz != sizeof(struct elf_prpsinfo)) {
				fprintf(stderr,
					"Size of prpsinfo differs (nhdr.n_descsz=%" \
					PRIu32 " vs. sizeof(struct elf_prpsinfo)=%lu)\n",
					nhdr.n_descsz,
					sizeof(struct elf_prpsinfo));
				continue;
			}

			memcpy(&core_prpsinfo, notes->d_buf + desc_pos,
			       sizeof(struct elf_prpsinfo));
#endif
			DEBUG("Found PRPSINFO (fname: %s, psargs: %s)\n",
			      core_fname, core_psargs);
			break;
		}
	}

	return 0;
}

static void core_parse_auxv(Elf *core, Elf_Data *data, size_t size,
			    GElf_Addr *auxv_at_phdr,
			    GElf_Addr *auxv_at_sysinfo_ehdr)
{
	GElf_auxv_t auxv_mem;
	GElf_auxv_t *auxv;
	size_t auxvent = size / gelf_fsize(core, ELF_T_AUXV, 1, EV_CURRENT);
	size_t i;

	GElf_Addr auxv_entry = 0;
	GElf_Addr auxv_phdr = 0;
	GElf_Xword auxv_phent = 0;
	GElf_Xword auxv_phnum = 0;
	size_t auxv_pagesz = 0;
	GElf_Addr interp_base = 0;
	GElf_Addr sysinfo_ehdr = 0;

	for (i = 0; i < auxvent; i++) {
		auxv = gelf_getauxv(data, i, &auxv_mem);
		if (!auxv) {
			abort();
		}

		GElf_Addr val = auxv->a_un.a_val;

		DEBUG("a_type=%" PRIu64 ", a_val=%#" PRIx64 "\n",
		      auxv->a_type, val);

		switch(auxv->a_type) {
		case AT_ENTRY:
			auxv_entry = val;
			break;
		case AT_PHDR:
			auxv_phdr = val;
			break;
		case AT_PHNUM:
			auxv_phnum = val;
			break;
		case AT_PHENT:
			auxv_phent = val;
			break;
		case AT_PAGESZ:
			auxv_pagesz = val;
			break;
		case AT_BASE:
			interp_base = val;
			break;
		case AT_SYSINFO_EHDR:
			sysinfo_ehdr = val;
			break;
		}
	}

	DEBUG("AUXV: phdr=%#" PRIx64 ", phnum=%" PRIu64 ", phent=%" PRIu64 \
	      ", base=%#" PRIx64 "\n",
	      auxv_phdr, auxv_phnum, auxv_phent, interp_base);
	*auxv_at_phdr = auxv_phdr;
	*auxv_at_sysinfo_ehdr = sysinfo_ehdr;
}

static Elf *map_search_ehdr(struct core_map_info *map)
{
	Elf_Data *data;
	unsigned char *e_ident;
	Elf *next;

	if (map->p_filesz < EI_NIDENT)
		return 0;
// ||
//	    !(phdr->p_type == PT_LOAD
//	      && ((phdr->p_flags & (PF_R|PF_X)) == (PF_R|PF_X))))

	/* Look for ELF header in this LOAD segment */
	data = elf_getdata_rawchunk (map->elf, map->p_offset,
				     map->p_filesz, ELF_T_EHDR);
	if (!data) {
		fprintf(stderr, "Couldn't read LOAD segment: %s\n",
			elf_errmsg(-1));
		return NULL;
	}
	e_ident = (unsigned char *)data->d_buf;
	if (memcmp (e_ident, ELFMAG, SELFMAG) != 0
	    || e_ident[EI_CLASS] <= ELFCLASSNONE
	    || e_ident[EI_CLASS] >= ELFCLASSNUM
	    || e_ident[EI_DATA] <= ELFDATANONE
	    || e_ident[EI_DATA] >= ELFDATANUM
	    || e_ident[EI_VERSION] <= EV_NONE
	    || e_ident[EI_VERSION] >= EV_NUM)
		return 0;

	DEBUG("Found another ELF header (%s, %s)\n",
	      elf_classnames[e_ident[EI_CLASS]],
	      elf_datanames[e_ident[EI_DATA]]);

	void *ehdr_mem = malloc(map->p_filesz);
	if (!ehdr_mem)
		abort();

	memcpy(ehdr_mem, data->d_buf, map->p_filesz);

	/* Hack against libelf test for e_shoff > p_filesz */
	if (gelf_getclass(map->elf) == ELFCLASS32) {
		Elf32_Ehdr *tmp = ehdr_mem;
		tmp->e_shoff = 0;
		tmp->e_shnum = 0;
	} else {
		Elf64_Ehdr *tmp = ehdr_mem;
		tmp->e_shoff = 0;
		tmp->e_shnum = 0;
	}

	next = elf_memory(ehdr_mem, map->p_filesz);
	if (next == NULL) {
		fprintf(stderr, "Error: %s\n", elf_errmsg (-1));
		return NULL;
	}

	return next;
}

#if 0
static GElf_Addr get_r_debug_symbol(GElf_Addr dyn_symtab, GElf_Word dyn_syment,
				    GElf_Addr dyn_strtab, GElf_Word dyn_strsz)
{
	struct core_map_info *map;
	GElf_Addr off;
	size_t size = 500;
	Elf_Data *data, *strdata;
	size_t i = 0;

	map = find_map(core_map, dyn_strtab, dyn_strsz);
	if (!map)
		return 0;

	off = dyn_strtab - map->p_vaddr;
	DEBUG("offset = %x, size = %x, maxsize = %x\n", off, size,
	      map->p_filesz - off - 1);
	strdata = elf_getdata_rawchunk (map->elf, map->p_offset + off,
					dyn_strsz, ELF_T_BYTE);
	if (!strdata) {
		fprintf(stderr, "Can't get dynamic string table: %s\n",
			elf_errmsg(-1));
		return 0;
	}

	map = find_map(core_map, dyn_symtab, size);
	if (!map)
		return 0;

	off = dyn_symtab - map->p_vaddr;
	DEBUG("offset = %x, size = %x, maxsize = %x\n", off, size,
	      map->p_filesz - off - 1);
	data = elf_getdata_rawchunk (map->elf, map->p_offset + off,
				     size,
				     ELF_T_SYM);
	if (!data) {
		fprintf(stderr, "Can't get DYNAMIC segment: %s\n",
			elf_errmsg(-1));
		return 0;
	}

	/* TODO */
	while (i < 100) {
		GElf_Sym sym;
		gelf_getsym(data, i, &sym);
		printf("%s\n", (char *)strdata->d_buf + sym.st_name);
		i++;
	}

	return 0;
}
#endif

void core_parse_dynamic(struct core_map_info *core_map, GElf_Addr dynamic,
			size_t size)
{
	struct core_map_info *map;
	Elf_Data *data;
	off_t off;
	GElf_Addr r_debug_addr;
	GElf_Dyn dyn_mem;
	size_t dynent;
	size_t i;

	map = find_map(core_map, dynamic, size);
	if (!map)
		return;

	off = dynamic - map->p_vaddr;
	DEBUG("offset = %x, size = %x, maxsize = %x\n", off, size,
	      map->p_filesz - off - 1);
	data = elf_getdata_rawchunk (map->elf, map->p_offset + off,
				     size,
				     ELF_T_DYN);
	if (!data) {
		fprintf(stderr, "Can't get DYNAMIC segment: %s\n",
			elf_errmsg(-1));
		return;
	}

	GElf_Addr symtab;
	GElf_Word syment;
	GElf_Addr strtab;
	GElf_Word strsz;

	dynent = size / gelf_fsize(map->elf, ELF_T_DYN, 1, EV_CURRENT);
	for (i = 0; i < dynent; i++) {
		GElf_Dyn *dyn = gelf_getdyn(data, i, &dyn_mem);
		if (!dyn)
			abort();

		DEBUG("Found d_tag=%" PRIu64 " d_val=%#" PRIx64 "\n",
		      dyn->d_tag, dyn->d_un.d_val);

		switch(dyn->d_tag) {
		case DT_DEBUG:
			DEBUG("Found r_debug at %#" PRIx64 "\n",
			      dyn->d_un.d_ptr);
			r_debug_addr = dyn->d_un.d_ptr;
			break;
		case DT_SYMTAB:
			DEBUG("Found DT_SYMTAB at %#" PRIx64 "\n",
			      dyn->d_un.d_ptr);
			symtab = dyn->d_un.d_ptr;
			break;
		case DT_SYMENT:
			DEBUG("Found DT_SYMENT with %#" PRIu64 "\n",
			      dyn->d_un.d_val);
			syment = dyn->d_un.d_val;
			break;
		case DT_STRTAB:
			DEBUG("Found DT_STRTAB at %#" PRIx64 "\n",
			      dyn->d_un.d_ptr);
			strtab = dyn->d_un.d_ptr;
			break;
		case DT_STRSZ:
			DEBUG("Found DT_STRSZ with %#" PRIu64 "\n",
			      dyn->d_un.d_val);
			strsz = dyn->d_un.d_val;
			break;
		case DT_NULL:
			goto out;
			break;
		}
	}
out:
	if (!r_debug_addr)
		return;

	map = find_map(core_map, r_debug_addr, sizeof(struct r_debug));
	if (!map)
		return;

	off = r_debug_addr - map->p_vaddr;
	DEBUG("offset = %x, size = %x, maxsize = %x\n", off, sizeof(struct r_debug),
	      map->p_filesz - off - 1);
	data = elf_getdata_rawchunk (map->elf, map->p_offset + off,
				     sizeof(struct r_debug),
				     ELF_T_BYTE);
	if (!data) {
		fprintf(stderr, "Can't get r_debug: %s\n", elf_errmsg(-1));
		return;
	}

	struct r_debug *r_debug = data->d_buf;
	GElf_Xword r_version;
	GElf_Addr r_map;
	if (gelf_getclass(map->elf) == ELFCLASS32) {
		r_version = *(uint32_t *) data->d_buf;

		memcpy(&r_version, r_debug, 4);
		DEBUG("linker protocol version %" PRIu32 "\n", r_version);
		r_map = *((uint32_t *)((char *)r_debug + 4));
	} else {
		memcpy(&r_version, r_debug, 8);
		DEBUG("linker protocol version %" PRIu64 "\n", r_version);
		r_map = *((uint64_t *)((char *)r_debug + 8));
	}

	if (r_version != 1) {
		fprintf(stderr, "Linker protocol version %" PRIu64	\
			" not supported\n", r_version);
		return;
	}

	struct Glink_map {
		GElf_Addr l_addr;
		uint64_t l_name;
		uint64_t l_ld;
		uint64_t l_next, l_prev;
	};


	DEBUG("Found r_map at %p\n", r_debug->r_map);
	DEBUG("Found r_map at %#" PRIx64 "\n", r_map);

	while (r_map) {
		struct core_map_info *map;
		struct Glink_map lmap_mem;
		struct Glink_map *lmap = &lmap_mem;

		/* Begin: gelf_getlinkmap() */
		map = find_map(core_map, r_map, sizeof(struct Glink_map));
		if (!map)
			return;

		off = r_map - map->p_vaddr;
		DEBUG("offset = %x, size = %x, maxsize = %x\n", off,
		      sizeof(struct Glink_map),
		      map->p_filesz - off - 1);
		data = elf_getdata_rawchunk (map->elf, map->p_offset + off,
					     sizeof(struct Glink_map),
					     ELF_T_BYTE);
		if (!data) {
			fprintf(stderr, "Can't get r_map: %s\n",
				elf_errmsg(-1));
			return;
		}

		if (gelf_getclass(map->elf) == ELFCLASS32) {
			struct link_map32 {
				Elf32_Addr l_addr;
				uint32_t l_name;
				uint32_t l_ld;
				uint32_t l_next, l_prev;
			};

			struct link_map32 *src = data->d_buf;

			lmap_mem.l_addr = src->l_addr;
			lmap_mem.l_name = src->l_name;
			lmap_mem.l_ld = src->l_ld;
			lmap_mem.l_next = src->l_next;
			lmap_mem.l_prev = src->l_prev;
		} else {
			lmap_mem = *((struct Glink_map *)data->d_buf);
		}

		DEBUG("r_map=%" PRIx64 "\n\tl_addr=%" PRIx64 "\n\tl_name=%#" \
		      PRIx64 "\n\tl_ld=%" PRIx64 "\n\tl_next=%" PRIx64 \
		      "\n\tl_prev=%"PRIx64"\n",
		      r_map, lmap->l_addr, lmap->l_name, lmap->l_ld,
		      lmap->l_next, lmap->l_prev);

#if 1
		char *l_name = malloc(100);
		if (!l_name)
			abort();
		get_map_data(core_map, lmap->l_name, 100, l_name);
		DEBUG("l_name = \"%s\"\n", l_name);
#endif

		/* Fill our own link map */
		struct link_map *mmap = malloc(sizeof(struct link_map));
		if (!mmap)
			abort();

		mmap->l_addr = lmap->l_addr;
		mmap->l_name = strdup(l_name);
		mmap->l_ld = (ElfW(Dyn) *)lmap->l_ld;
		mmap->l_next = core_link_map;
		mmap->l_prev = NULL;
		if (core_link_map)
			core_link_map->l_prev = mmap;
		core_link_map = mmap;

		r_map = lmap->l_next;
	}
}

static GElf_Addr get_vdso_link_map_l_addr(struct core_map_info *core_map,
					  GElf_Addr core_sysinfo_ehdr)
{
	struct core_map_info *map;
	Elf *elf;
	GElf_Ehdr ehdr;
	int i;
	GElf_Addr vdso_l_map_start;
	GElf_Addr vdso_l_map_end;
	GElf_Addr vdso_l_addr;
	GElf_Addr vdso_l_ld;

	map = find_map(core_map, core_sysinfo_ehdr, ELF_T_EHDR);
	if (!map) {
		fprintf(stderr, "Couldn't map the vDSO referenced in AUXV\n");
		return 0;
	}

	elf = map_search_ehdr(map);
	if (!elf) {
		fprintf(stderr, "Couldn't find the ELF header of the vDSO\n");
		return 0;
	}

	if (gelf_getehdr (elf, &ehdr) == NULL) {
		fprintf(stderr, "Couldn't get the ELF header: %s\n",
			elf_errmsg (-1));
		return 0;
	}

	/* Compute the load address of dynamic linked executable */
	for (i = 0; i < ehdr.e_phnum; i++) {
		GElf_Phdr phdr_mem;
		GElf_Phdr *phdr = gelf_getphdr (elf, i, &phdr_mem);

		switch (phdr->p_type) {
		case PT_LOAD:
			vdso_l_addr = phdr->p_vaddr;
			vdso_l_map_end = phdr->p_vaddr +
				phdr->p_memsz;
			break;
		case PT_DYNAMIC:
			vdso_l_ld = phdr->p_vaddr;
			break;
		case PT_TLS:
			/* There must be no TLS segment.  */
			abort();
			break;
		}
	}

	vdso_l_map_start = core_sysinfo_ehdr;
	vdso_l_addr = vdso_l_map_start - vdso_l_addr;
	vdso_l_map_end += vdso_l_addr;
	vdso_l_ld = vdso_l_ld + vdso_l_addr;
	DEBUG("vDSO: l_addr=%#lx, l_ld=%#lx, l_map_end=%#lx\n",
	      vdso_l_addr, vdso_l_ld, vdso_l_map_end);

	return vdso_l_addr;
}

static int process_core(Elf *elf, GElf_Ehdr *ehdr)
{
	struct core_map_info *core_map = NULL;
	int i;

	if (print_debug)
		fprintf(stderr,
			"idx type    %*s %*s %*s %*s %*s  align flags\n",
			gelf_getclass (elf) == ELFCLASS32 ? 9 : 17, "offset",
			gelf_getclass (elf) == ELFCLASS32 ? 10 : 18, "vaddr",
			gelf_getclass (elf) == ELFCLASS32 ? 10 : 18, "paddr",
			gelf_getclass (elf) == ELFCLASS32 ? 9 : 12, "filesz",
			gelf_getclass (elf) == ELFCLASS32 ? 9 : 12, "memsz");

	for (i = 0; i < ehdr->e_phnum; i++) {
		GElf_Phdr mem;
		GElf_Phdr *phdr = gelf_getphdr (elf, i, &mem);

		if (print_debug)
			print_phdr(elf, phdr, i, 0);

		switch (phdr->p_type) {
		case PT_NOTE:
			if (core_handle_note(elf, phdr))
				return -1;
			break;
		case PT_LOAD:
			core_add_map_info(&core_map, elf, phdr);
			break;
		}
	}

	if (auxv && auxv_size != 0) {
		GElf_Addr load_addr;
		GElf_Addr auxv_phdr;
		GElf_Addr auxv_sysinfo_ehdr;
		struct core_map_info *map;
		Elf *next;

		core_parse_auxv(elf, auxv, auxv_size, &auxv_phdr,
				&auxv_sysinfo_ehdr);
		DEBUG("Found program header at %p\n", auxv_phdr);
		map = find_map(core_map, auxv_phdr, ELF_T_EHDR);
		if (!map) {
			fprintf(stderr,
				"Couldn't map the program header referenced in AUXV\n");
			return -1;
		}
		next = map_search_ehdr(map);
		if (!next) {
			fprintf(stderr,
				"Couldn't find the ELF header of the executable\n");
			return -1;
		}

		/* Get the Applications ELF header */
		GElf_Ehdr next_ehdr;
		if (gelf_getehdr (next, &next_ehdr) == NULL) {
			fprintf(stderr, "Couldn't get the ELF header: %s\n",
				elf_errmsg (-1));
			return -1;
		}

		/* Compute the load address of dynamic linked executable */
		do {
			GElf_Phdr phdr_mem;
			GElf_Phdr *phdr = gelf_getphdr (next, 0, &phdr_mem);

			if (phdr->p_type != PT_PHDR) {
				DEBUG("Couldn't find the PHDR segment, probably statically linked executable\n");
			} else {
				load_addr = auxv_phdr - phdr->p_vaddr;
				DEBUG("Load Address at %#" PRIx64 " (AUXV=%#" \
				      PRIx64 ",PHDR=%#" PRIx64")\n",
				      load_addr, auxv_phdr, phdr->p_vaddr);
				core_l_addr = auxv_phdr - phdr->p_offset;
			}
		} while(0);

		/* Look for NOTE segments that contain the build-id note */
		int j;
		for (j = 0; j < next_ehdr.e_phnum; ++j) {
			GElf_Phdr nmem;
			GElf_Phdr *nphdr = gelf_getphdr (next, j, &nmem);

			if (print_debug)
				print_phdr(next, nphdr, j, 2);

			/* Read the build-id */
			switch (nphdr->p_type) {
			case PT_NOTE:
			{
				unsigned char build_id[nphdr->p_filesz];
				size_t len = get_buildid(next, nphdr, build_id,
							 nphdr->p_filesz);
				if (len) {
					unsigned int i;
					printf("%s %#" PRIx64 " ",
					       core_fname ? core_fname : "",
					       core_l_addr);
					for (i = 0; i < len; i++)
						printf("%02" PRIx8,
						       build_id[i]);
					printf("\n");
				}
			}
				break;
			case PT_DYNAMIC:
				core_parse_dynamic(core_map,
						   load_addr + nphdr->p_vaddr,
						   nphdr->p_memsz);
				break;
			}
		}

		/* Print the build-id of the vDSO */
		if (print_vdso_buildid && auxv_sysinfo_ehdr)
			so_print_buildid(core_map, "vDSO", auxv_sysinfo_ehdr);

		/* Compute load address of vDSO in link map */
		if (auxv_sysinfo_ehdr)
			core_vdso_l_addr = get_vdso_link_map_l_addr(core_map,
								    auxv_sysinfo_ehdr);

		/* Print build-id of shared objects */
		struct link_map *tmp;
		for (tmp = core_link_map; tmp; tmp = tmp->l_next) {
			struct core_map_info *so_map_info;
			Elf *so_elf;

			/* We already printed this build-id */
			if ((tmp->l_addr == core_l_addr) ||
			    (tmp->l_addr == core_vdso_l_addr))
				continue;

			so_print_buildid(core_map, tmp->l_name, tmp->l_addr);
		}
	}

	return 0;
}

static int process_executable(Elf *elf, GElf_Ehdr *ehdr)
{
	int i, rc, found = 0;
	Elf_Scn *scn = NULL;
	size_t shstrndx;

	DEBUG("Parse Program Header Table\n");

	if (print_debug)
		fprintf(stderr,
			"idx type    %*s %*s %*s %*s %*s  align flags\n",
			gelf_getclass (elf) == ELFCLASS32 ? 9 : 17, "offset",
			gelf_getclass (elf) == ELFCLASS32 ? 10 : 18, "vaddr",
			gelf_getclass (elf) == ELFCLASS32 ? 10 : 18, "paddr",
			gelf_getclass (elf) == ELFCLASS32 ? 9 : 12, "filesz",
			gelf_getclass (elf) == ELFCLASS32 ? 9 : 12, "memsz");

	for (i = 0; i < ehdr->e_phnum; i++) {
		unsigned char build_id[20];
		GElf_Phdr mem;
		GElf_Phdr *phdr = gelf_getphdr (elf, i, &mem);

		if (print_debug)
			print_phdr(elf, phdr, i, 0);

		switch (phdr->p_type) {
		case PT_NOTE:
		{
			size_t len = get_buildid(elf, phdr, build_id, 20);
			if (len) {
				unsigned int i;
				printf("%s ",
				       core_fname ? core_fname : "");
				for (i = 0; i < len; i++)
					printf("%02" PRIx8, build_id[i]);
				printf("\n");
				found = 1;
			}
			break;
		}
		}
	}

	if (found)
		return 0;

	DEBUG("Parse Section Header Table\n");

	if (print_debug) {
		rc = elf_getshstrndx(elf, &shstrndx);
		if (rc) {
			fprintf(stderr, "Couldn't get .shstrtab index: %s\n",
				elf_errmsg(-1));
			return -1;
		}
	}

	while ((scn = elf_nextscn(elf, scn)) != NULL) {
		GElf_Shdr shdr_mem;
		GElf_Shdr *shdr = gelf_getshdr(scn, &shdr_mem);

		if (!shdr) {
			fprintf(stderr, "Couldn't get section header: %s\n",
				elf_errmsg(-1));
			return -1;
		}

		if (print_debug) {
			char *name;

			name = elf_strptr(elf, shstrndx, shdr->sh_name);
			if (!name)
				fprintf(stderr, "Couldn't read section name: %s\n",
					elf_errmsg(-1));
			fprintf(stderr, "DEBUG: section \"%s\"\n",
				name ? name : "");
		}

		switch (shdr->sh_type) {
		case SHT_NOTE: {
			Elf_Data *data = elf_getdata(scn, NULL);
			if (!data) {
				fprintf(stderr, "Couldn't get section data: %s\n",
					elf_errmsg(-1));
				return -1;
			}

			unsigned char dest[shdr->sh_size];
			size_t dest_len = shdr->sh_size;

			rc = nhdr_getbuildid(data, dest, dest_len);
			if (rc > 1) {
				unsigned int i;
				printf("%s ",
				       core_fname ? core_fname : "");
				for (i = 0; i < rc; i++)
					printf("%02" PRIx8, dest[i]);
				printf("\n");
				found = 1;
			}
		}
		}
	}

	return 0;
}

int main (int argc, char *argv[])
{
	poptContext optCon;   /* context for parsing command-line options */
	const char *arg;
	int rc;

	optCon = poptGetContext("readcore", argc, (const char **)argv,
				optionsTable, 0);

	while ((rc = poptGetNextOpt (optCon)) > 0) {
		switch(rc) {
		case POPT_ERROR_BADOPT:
			/* do nothing */ ;
			break;
		}
	}

	if (print_kernel_buildid && !poptGetArg(optCon)) {
		DEBUG("Processing kernel note file\n");
		rc = kernel_print_buildid();
		return rc;
	}

	arg = poptGetArg(optCon);
	if (!arg) {
		poptPrintHelp(optCon, stdout, 0);
		return 1;
	}

	/* Open the file.  */
	int fd = open (arg, O_RDONLY);
	if (fd == -1) {
		error (0, errno, "cannot open input file");
		return -1;
	}

	struct stat stat_buf;
	rc = fstat(fd, &stat_buf);
	if (rc) {
		error (0, errno, "cannot stat input file");
		goto closefd;
	}

	if (!S_ISREG(stat_buf.st_mode)) {
		error (0, 0, "cannot open input file: Not a regular file");
		goto closefd;
	}

	/* Before we start tell the ELF library which version we are using.  */
	elf_version (EV_CURRENT);

	/* Create an `Elf' descriptor.  */
	Elf *elf = elf_begin (fd, ELF_C_READ_MMAP, NULL);
	if (elf == NULL) {
		error (0, 0, elf_errmsg (-1));
		rc = -1;
		goto closefd;
	}

	if (elf_kind(elf) != ELF_K_ELF) {
		fprintf(stderr, "This is not an ELF file\n");
		rc = -1;
		goto out;
	}

	DEBUG("This is an ELF file\n");

	GElf_Ehdr ehdr_mem, *ehdr;
	if ((ehdr = gelf_getehdr (elf, &ehdr_mem)) == NULL) {
		fprintf(stderr, "Can't get the ELF header: %s\n",
			elf_errmsg (-1));
		rc = -1;
		goto out;
	}

	switch (ehdr->e_type) {
	case ET_CORE:
		DEBUG("Processing core file\n");
		rc = process_core(elf, ehdr);
		break;
	case ET_REL:
	case ET_DYN:
	case ET_EXEC:
	{
		DEBUG("Processing executable file\n");
		const char *index = strrchr(arg, '/');
		if (index)
			core_fname = index + 1;
		else
			core_fname = arg;
		rc = process_executable(elf, ehdr);
		break;
	}
	default:
		break;
	}

out:
	/* Now we can close the descriptor.  */
	if (elf_end (elf) != 0)
		error (0, 0, elf_errmsg (-1));

closefd:
	close (fd);
	return rc;
}
